/***
   NAME
     escbox.h
   PURPOSE
     This file contains all include statements for the program, some
     constant and type definitions that are used in the population data
     representations and the prototypes of various functions.
     This file is meant to be included in all escalator boxcar train
     modules, as it contains the global definitions that are accessible to
     all modules.
   NOTES

   HISTORY
     AMdR - Jul 19, 1995 : Created.
     AMdR - Oct 25, 2012 : Revised last.
***/

#ifndef ESCBOX_H
#define ESCBOX_H

#if defined(MODULE)
#include "ebttooldefs.h"
#endif

#define	RK2	98110801			/* Key values to indicate   */
#define RK4	98110802			/* the different            */
#define	RKF45	98110803			/* integration methods      */
#define	RKCK	98110804
#define	DOPRI5	98110805
#define	DOPRI8	98110806
#define	RADAU5	98110807

#define MAXDERS	20				/* Max. stages in ODE solver*/

#include "cr.h"
//#if defined(PROBLEMFILE)                        /* If program header file   */
//#include PROBLEMFILE                            /* is defined, include it.  */
//#elif ANSI_C
//#error You failed to specify the PROBLEMFILE constant. Use the "ebt" script!
//#endif

#ifndef TIME_METHOD
#define	TIME_METHOD		RKCK
#endif

#ifndef DYNAMIC_COHORTS
#define DYNAMIC_COHORTS		0
#endif
#if (DYNAMIC_COHORTS > 0)
#define DYNAMIC_COHORTS		1
#if ((TIME_METHOD !=  DOPRI5) && (TIME_METHOD !=  DOPRI8) && \
     (TIME_METHOD !=  RADAU5))
#undef 	TIME_METHOD
#define TIME_METHOD		DOPRI5
#endif
#else
#define DYNAMIC_COHORTS		0
#endif

#ifndef EVENT_NR
#define EVENT_NR		0
#endif
#if (EVENT_NR > 0)
#if ((TIME_METHOD !=  DOPRI5) && (TIME_METHOD !=  DOPRI8) && \
     (TIME_METHOD !=  RADAU5))
#undef 	TIME_METHOD
#define TIME_METHOD		DOPRI5
#endif
#else
#define EVENT_NR		0
#endif

#ifndef BIFURCATION
#define BIFURCATION		0
#ifndef MEASUREBIFSTATS
#define MEASUREBIFSTATS		0
#endif
#elif (BIFURCATION == 1)
#ifndef MEASUREBIFSTATS
#define MEASUREBIFSTATS		1
#endif
#endif

#ifndef CHECK_EXTINCTION
#define CHECK_EXTINCTION	2			// 0: Ignore all tests; 1: Ignore run ending; 2: Test and end run
#endif

#include  "ebttune.h"
#include  "ctype.h"
#include  "math.h"
#include  "stdio.h"
#include  "string.h"

#if HAS_FLOAT_H
#include  "float.h"
#endif

#if HAS_SIGNALS
#include  "signal.h"
#endif

#if defined(_GNU_SOURCE)
#include  "fenv.h"
#endif

#if ANSI_C
#include  "limits.h"
#include  "stddef.h"
#include  "stdlib.h"
#include  "stdarg.h"
#include  "sys/stat.h"
#else
#include  "malloc.h"
#include  "memory.h"
#endif /* ANSI_C */


/*==========================================================================*/
/*
 * Definition of some global constants. Most of these definitions only take
 * effect if the flag EBTLIB is defined, which occurs in the
 * Escalator Boxcar Train module files.
 */

#if (POPULATION_NR > 0)				/* The number of replicate  */
#define   COHORT_SIZE    (1+I_STATE_DIM)	/* size of cohort           */
#else
#define   I_CONST_DIM    0
#define   I_STATE_DIM    0
#define   COHORT_SIZE    0
#endif

#if defined(DBL_MAX)				/* Stub value for missing   */
#define MISSING_VALUE		(DBL_MAX)	/* data point	            */
#elif defined(MAXDOUBLE)
#define MISSING_VALUE		(MAXDOUBLE)
#else
#define MISSING_VALUE		(1.23456789e+307)
#endif

#define   BIFTINY	 1.0E-4			/* Tiny value used in       */
						/* bifurcation computations */
#ifdef EBTLIB
						/* Maximum length of the    */
#if defined(PATH_MAX)				/* full path to a file      */
#define MAXFILENAMELEN	 	(PATH_MAX)
#elif defined(MAXPATHLEN)
#define MAXFILENAMELEN	 	(MAXPATHLEN)
#elif defined(_MAX_PATH)
#define MAXFILENAMELEN	 	(_MAX_PATH)
#else
#define MAXFILENAMELEN	 	(512)
#endif
						/* The maximum length of    */
#define   DESCRIP_MAX    80                     /* a quantity description   */

						/* The maximum length of    */
#define   REPORTNOTE_MAX 4096                   /* a ReportNote             */

#define	  MEM_BLOCK_SIZE 256			/* Number of doubles in     */
						/* allocated memory block   */
#define   SMALLEST_STEP  1.0E-12		/* The minimum step size    */
						/* allowed in integration   */
#define   DEFAULT_STEP	 0.1			/* Default step size in     */
						/* integration              */
#define   LARGEST_STEP   1.0			/* The minimum step size    */
						/* allowed in integration   */
#define	  MIN_ACCURACY	 1.0E-16		/* The minimum accuracy     */
						/* allowed in integration   */
#endif /* EBTLIB */

#define   NOSWITCH       -1			/* Stub values for problem  */
#define   NO_EVENT       -1			/* without events           */
#define   NO_COHORT_END  0
#define   COHORT_END     1


#ifndef I_CONST_DIM
#define I_CONST_DIM	0
#endif


/*==========================================================================*/
/*
 * Definition of the cohort structure that will hold the data of one
 * population cohort. Also defined is a pointer type to a single cohort and
 * the population data type.
 */

typedef double		cohort[COHORT_SIZE];
typedef	cohort		*cohort_pnt;
typedef	cohort_pnt	population;
#if (I_CONST_DIM > 0)
typedef double		cohortID[I_CONST_DIM];
typedef	cohortID	*cohortID_pnt;
typedef	cohortID_pnt	popID;
#endif
typedef int		cohort_ind[COHORT_SIZE];
typedef	cohort_ind	*cohort_ind_pnt;
typedef	cohort_ind_pnt	population_index;



/*==========================================================================*/
/*
 * Defining macro's for global use in all files
 */

#define number			(0)
#define i_state(a)		((a+1))
#define i_const(a)		((a))

#if ANSI_C
#define	DEF_TYPE		void		/* Basic pointer type       */
#define CONST			const
#define SIZE_TYPE		size_t
#define PROTO1(a)		a
#define PROTO2(a,b)		a,b
#define PROTO3(a,b,c)		a,b,c
#define PROTO4(a,b,c,d)		a,b,c,d
#define PROTO5(a,b,c,d,e)	a,b,c,d,e
#define PROTO6(a,b,c,d,e,f)	a,b,c,d,e,f
#define PROTO7(a,b,c,d,e,f,g)	a,b,c,d,e,f,g
#else
#define	DEF_TYPE		char
#define CONST
#define SIZE_TYPE		unsigned
#define PROTO1(a)
#define PROTO2(a,b)
#define PROTO3(a,b,c)
#define PROTO4(a,b,c,d)
#define PROTO5(a,b,c,d,e)
#define PROTO6(a,b,c,d,e,f)
#define PROTO7(a,b,c,d,e,f,g)
#endif /* ANSI_C */

#define MemBlocks(a)		(((a/MEM_BLOCK_SIZE)+1)*MEM_BLOCK_SIZE)
#define EBTDEBUG(a)		(dbgfil && (debug_level >= (a)))



/*==========================================================================*/
/*
 * Definition of global variables and functions accessible in user
 * specified problem file.
 */

#ifndef EBTLIB
#define CycleStart	SetBpointNo
#define CycleEnd	InstantDynamics

extern double		cohort_limit;
extern double		next_output, next_state_output;
#if (POPULATION_NR > 0)
extern int		cohort_no[POPULATION_NR];
extern int		bpoint_no[POPULATION_NR];
#endif // (POPULATION_NR > 0)
extern int		rk_level;
extern int		ForcedCohortEnd;
extern int		ForcedRunEnd;
extern int		LocatedEvent;
extern int		parameter_nr;
#if PARAMETER_NR
extern double		parameter[PARAMETER_NR];
#endif
#if (BIFURCATION == 1)
extern int		BifParIndex;
extern double		BifOutput;
extern double		BifStateOutput;
extern double		BifPeriod;
#endif // (BIFURCATION == 1)
#if ((POPULATION_NR > 0) && (I_CONST_DIM > 0))
extern popID		popIDcard[POPULATION_NR];
extern popID		ofsIDcard[POPULATION_NR];
#endif
extern int		AddCohorts(PROTO3(population *, int, int));

extern int		imax(PROTO2(int, int));
extern int		imin(PROTO2(int, int));
extern double		max(PROTO2(double, double));
extern double		min(PROTO2(double, double));

extern int		iszero(PROTO1(double));
extern int		isequal(PROTO2(double, double));
extern void		SetStepSize(PROTO1(double));
extern void		SievePop(PROTO1(void));
extern void		ErrorAbort(PROTO1(const char *));
extern void		ErrorExit(PROTO2(const int, const char *));
extern void		Warning(PROTO1(const char *));
#if ANSI_C
extern void		ReportNote(PROTO2(const char *, ...));
extern void		LabelState(PROTO3(int, const char *, ...));
extern void		measureBifstats(PROTO4(double *env, population *pop, int Nstores, ...));
#endif /* ANSI_C */
#endif /* EBTLIB */




/*==========================================================================*/
/*
 * Prototyping all functions that are defined in the problem-specific
 * program file
 */

#ifdef	EBTLIB
#undef	EXTERN
#define EXTERN	extern
#else
#undef	EXTERN
#define EXTERN
#endif
EXTERN void	UserInit(PROTO4(int, char **, double *, population *));
EXTERN void	SetBpointNo(PROTO3(double *, population *, int *));
EXTERN void	SetBpoints(PROTO3(double *, population *, population *));
EXTERN void	Gradient(PROTO7(double *, population *, population *,
				double *, population *, population *,
				population *));
EXTERN void	EventLocation(PROTO5(double *, population *, population *,
				     population *, double *));
EXTERN int	ForceCohortEnd(PROTO4(double *, population *, population *,
				      population *));
EXTERN void	InstantDynamics(PROTO3(double *, population *, population *));
EXTERN void	DefineOutput(PROTO3(double *, population *, double *));



/*==========================================================================*/
#endif /* ESCBOX_H */
